#!/usr/bin/env python
"""
Proof-of-Concept Implementation of a "Preloading Read-Only File System"
(C) 2009 Martin J. Fiedler 
This software is public domain.
"""
import sys, os, stat, errno, zlib, cPickle, cStringIO, thread, threading, time
try:
    import fuse
    fuse.fuse_python_api = (0, 2)
except ImportError:
    print "Fatal Error: FUSE Python bindings not installed!"
    sys.exit(1)

DEFAULT_BLOCKSIZE = 4096   # data file block size
ARTIFICIAL_SLOWNESS = 0.5  # time to wait between loading blocks (0 = no delay)

class FSObject(object):
    """A class representing an object (file, directory, ...) of the file
    system."""
    def __init__(self, name, **kwargs):
        """Initialize the structure. The name must be specified, everything
        else can be specified using keyword arguments."""
        self.name = name
        self.inode = 0
        self.mode = 0
        self.uid = 0
        self.gid = 0
        self.atime = 0
        self.mtime = 0
        self.ctime = 0
        self.rdev = 0
        self.size = 0
        self.csize = 0
        self.offset = 0
        self.__dict__.update(kwargs)
    def isfile(self):
        return stat.S_ISREG(self.mode)
    def isdir(self):
        return stat.S_ISDIR(self.mode)
    def islink(self):
        return stat.S_ISLNK(self.mode)
    def isdev(self):
        return stat.S_ISCHR(self.mode) or stat.S_ISBLK(self.mode)
    def isfifo(self):
        return stat.S_ISFIFO(self.mode)
    def issock(self):
        return stat.S_ISSOCK(self.mode)
    def getobj(self, name):
        """If this is a directory, return the entry with the specified name."""
        if not hasattr(self, 'objs'):
            raise KeyError
        if not hasattr(self, 'index'):
            self.index = dict([(self.objs[i].name, i) for i in xrange(len(self.objs))])
        try:
            return self.objs[self.index[name]]
        except IndexError:
            raise KeyError
    def __repr__(self):
        keys = [k for k in self.__dict__ if not k.startswith('_')]
        keys.remove('name')
        return "%s(%r, %s)" % (self.__class__.__name__, self.name, ", ".join(["%s=%s" % (k, repr(self.__dict__[k])) for k in keys]))

def IOError_with_errno(code):
    """Raise an IOError with a specified error number. Used to signal errors
    to FUSE."""
    e = IOError()
    e.errno = code
    return e

class ProFS(fuse.Fuse):
    """The main class of the Preloading Read-Only File System."""

    def __init__(self, *args, **kwargs):
        fuse.Fuse.__init__(self, *args, **kwargs)
        self.blocksize = DEFAULT_BLOCKSIZE
        self.files = {}
        self.buflock = thread.allocate_lock()
        self.wakeup = threading.Event()

    def fsinit(self):
        """Open and parse the index file, open the data file and start the
        background reading thread."""
        try:
            infile = self.cmdline[1][0]
        except IndexError:
            print "error: input file not specified"
            sys.exit(2)
        # determine extension of the files (upper- or lowercase)
        ext = os.path.splitext(os.path.basename(infile))[-1]
        idx_ext = ".idx"
        dat_ext = ".dat"
        if ext.lower() in (".", ".idx", ".dat"):
            infile = infile[:-4]
            if ext.upper() == ext:
                idx_ext = ".IDX"
                dat_ext = ".DAT"
        # read the index file
        try:
            f = open(infile + idx_ext, "rb")
            stream = cStringIO.StringIO(zlib.decompress(f.read()))
            f.close()
            self.root = cPickle.load(stream)
            del stream
        except IOError, e:
            print "error reading the index file:", e
            sys.exit(1)
        # open the data file
        try:
            self.datafile = open(infile + dat_ext, "rb")
            self.datafile.seek(0, 2)
            blockcount = (self.datafile.tell() + self.blocksize - 1) / self.blocksize
        except IOError, e:
            print "error opening the data file:", e
            sys.exit(1)
        # open the log file
        if self.cmdline[0].logfile:
            try:
                self.logfile = open(self.cmdline[0].logfile, "w")
            except IOError, e:
                print "error opening the log file:", e
                self.logfile = None
        else:
            self.logfile = None
        self.logged = set()
        # start the loader thread
        self.blocks = [None] * blockcount
        self.requests = range(blockcount)
        thread.start_new_thread(self.loader, tuple())

    def loader(self):
        """The background reading thread. Reads blocks from the data file into
        memory until all blocks have been loaded."""
        while True:
            self.buflock.acquire()
            try:
                if not self.requests:
                    return
                block = self.requests.pop(0)
            finally:
                self.buflock.release()
            if self.blocks[block]:
                continue  # block already loaded
            try:
                self.datafile.seek(block * self.blocksize)
                data = self.datafile.read(self.blocksize)
                time.sleep(ARTIFICIAL_SLOWNESS)
                self.buflock.acquire()
                try:
                    self.blocks[block] = data
                finally:
                    self.buflock.release()
                self.wakeup.set()
            except IOError, e:
                print "IO error in block %d:" % block, e
                self.blocks[block] = "XXX"  # just a dummy so we don't try to
                                            # load that block again

    def getobj(self, path):
        """Retrieves the FSObject instance for a specific path or raises an
        IOError with the appropriate errno in case of failure."""
        obj = self.root
        for component in path.split('/'):
            if not component:
                continue  # ignore double-slashes
            if not obj.isdir():
                raise IOError_with_errno(errno.ENOTDIR)
            try:
                obj = obj.getobj(component)
            except (KeyError, IndexError):
                raise IOError_with_errno(errno.ENOENT)
        return obj

    def obj2block(self, obj):
        """Returns the list of buffer blocks occupied by a specific FSObject."""
        if not(obj.isfile()) or not(obj.csize):
            return []
        return range(obj.offset / self.blocksize, (obj.offset + obj.csize - 1) / self.blocksize + 1)

    def need(self, blocklist):
        """Returns a list of all buffer blocks that are also part of the
        specified blocklist and that have not yet been loeaded into the
        buffer."""
        self.buflock.acquire()
        try:
            return [b for b in blocklist if not self.blocks[b]]
        finally:
            self.buflock.release()

    def prefetch(self, blocklist):
        """Makes sure that the specified blocks are loaded as soon as
        possible, if they aren't already."""
        need = self.need(blocklist)
        if need:
            self.buflock.acquire()
            try:
                self.requests = need + self.requests
            finally:
                self.buflock.release()
        return len(need)

    def getfile(self, path):
        """Retrieves the contents of a file. If the file is not already open,
        wait until all blocks of the compressed image of the file have been
        loaded, decompress the file, add it to the cache and return the
        contents. If the file is already open, it means that it's in the
        cache and can be returned directly."""
        try:
            return self.files[path]
        except KeyError:
            pass
        obj = self.getobj(path)
        if not(obj.isfile()):
            raise IOError_with_errno(errno.EISDIR)
        blockrange = self.obj2block(obj)
        if self.prefetch(blockrange):
            # Some blocks have been prefetched, that means we need to wait
            # for these blocks. Just wait for the background loader thread
            # to load the next block and see if that solves our problem. If
            # not, wait again.
            while self.need(blockrange):
                self.wakeup.wait()
                self.wakeup.clear()
        block = obj.offset / self.blocksize
        offset = obj.offset % self.blocksize
        length = obj.csize
        data = ""
        z = zlib.decompressobj()
        while length:
            sub = self.blocks[block][offset:(offset+length)]
            data += z.decompress(sub)
            block += 1
            offset = 0
            length -= len(sub)
        data += z.flush()
        self.files[path] = data
        return data

    def getattr(self, path):
        """FUSE getattr() call: Retrieves information about a filesystem
        object (file or directory)."""
        obj = self.getobj(path)
        return fuse.Stat(
            st_ino=obj.inode,
            st_dev=0,
            st_blksize=4096,
            st_nlink=1,
            st_rdev=obj.rdev,
            st_uid=obj.uid,
            st_gid=obj.gid,
            st_atime=obj.atime,
            st_mtime=obj.mtime,
            st_ctime=obj.ctime,
            st_mode=obj.mode,
            st_size=obj.size,
            st_blocks=(obj.size + 4095) & (~4095))

    def readdir(self, path, offset=None):
        """FUSE readdir() call: Returns all file names in a directory."""
        obj = self.getobj(path)
        if not obj.isdir():
            raise IOError_with_errno(errno.ENOTDIR)
        return [fuse.Direntry(".", type=stat.S_IFDIR), fuse.Direntry("..", type=stat.S_IFDIR)] \
             + [fuse.Direntry(o.name) for o in obj.objs]

    def open(self, path, flags):
        """FUSE open() call: Opens a file. We don't need to do much here
        except checking if nobody tries to open a file for writing. We
        also write the filename into the log file and prefetch the file
        into memory, but we don't wait for it to arrive just yet --
        this will be done in the first read() call instead."""
        if (flags & (os.O_WRONLY | os.O_RDWR | os.O_CREAT | os.O_TRUNC)):
            raise IOError_with_errno(errno.EROFS)
        obj = self.getobj(path)
        self.prefetch(self.obj2block(obj))
        obj.atime = time.time()
        if self.logfile and not(path in self.logged):
            self.logfile.write(path + "\n")
            self.logfile.flush()
            self.logged.add(path)

    def read(self, path, size, offset, fd=None):
        """FUSE read() call: Reads a block of data from a file. The main
        work is done in getfile()."""
        return self.getfile(path)[offset:(offset+size)]

    def release(self, path, fd=None):
        """FUSE release() call: Closes the last instance of a file. This
        is where we remove the file from the decompressed files cache."""
        try:
            del self.files[path]
        except KeyError:
            pass

    def flush(self, path, fd=None):
        """FUSE flush() call: Not required for this filesystem."""
        pass

    def readlink(self, path):
        """FUSE readlink() call: Retrieves the path a symbolic link points
        to."""
        obj = self.getobj(path)
        if not obj.islink():
            raise IOError_with_errno(errno.EINVAL)
        return obj.link

    def mknod(self, path, mode, rdev=None):
        """FUSE mknod() call: Creates a file. Not implemented because this is
        a read-only filesystem."""
        raise IOError_with_errno(errno.EROFS)

    def mkdir(self, path, mode):
        """FUSE mkdir() call: Creates a file. Not implemented because this is
        a read-only filesystem."""
        raise IOError_with_errno(errno.EROFS)


# main program -> instantiate and run the filesystem
if __name__ == "__main__":
    fs = ProFS(dash_s_do='setsingle')
    fs.parser.add_option("-l", "--log", metavar="LOGFILE", dest="logfile",
                         help="access log file path")
    fs.parse()
    fs.main()
